@@ -22,10 +22,21 @@
 enum {
   TAG_COL_KEY = 0,
   TAG_COL_VALUE,
+  TAG_COL_COLLISION,
   TAG_COL_DATA,
   TAG_NUM_COLS
 };
 
+gboolean info_tag_key_collision(tag_t *tags, tag_t *tag) {
+  while(tags) {
+    if((tags != tag) && (strcasecmp(tags->key, tag->key) == 0))
+      return TRUE;
+
+    tags = tags->next;
+  }
+  return FALSE;
+}
+
 static gboolean
 view_selection_func(GtkTreeSelection *selection, GtkTreeModel *model,
 		     GtkTreePath *path, gboolean path_currently_selected,
@@ -42,12 +53,14 @@ view_selection_func(GtkTreeSelection *selection, GtkTreeModel *model,
     if(context->but_remove && context->but_edit) {
       
       /* you just cannot delete or edit the "created_by" tag */
-      if(strcasecmp(tag->key, "created_by") == 0) {
-	gtk_widget_set_sensitive(context->but_remove, FALSE);
-	gtk_widget_set_sensitive(context->but_edit, FALSE);
-      } else {
-	gtk_widget_set_sensitive(context->but_remove, TRUE);
-	gtk_widget_set_sensitive(context->but_edit, TRUE);
+      if(context->but_remove && context->but_edit) {
+	if(strcasecmp(tag->key, "created_by") == 0) {
+	  gtk_widget_set_sensitive(context->but_remove, FALSE);
+	  gtk_widget_set_sensitive(context->but_edit, FALSE);
+	} else {
+	  gtk_widget_set_sensitive(context->but_remove, TRUE);
+	  gtk_widget_set_sensitive(context->but_edit, TRUE);
+	}
       }
     }
   }
@@ -55,6 +68,27 @@ view_selection_func(GtkTreeSelection *selection, GtkTreeModel *model,
   return TRUE; /* allow selection state to change */
 }
 
+static void update_collisions(GtkListStore *store, tag_t *tags) {
+  GtkTreeIter iter;
+  tag_t *tag = NULL;
+      
+  /* walk the entire store to get all values */
+  if(gtk_tree_model_get_iter_first(GTK_TREE_MODEL(store), &iter)) {
+    gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, TAG_COL_DATA, &tag, -1);
+    g_assert(tag);      
+    gtk_list_store_set(store, &iter,
+       TAG_COL_COLLISION, info_tag_key_collision(tags, tag), -1);
+
+    while(gtk_tree_model_iter_next(GTK_TREE_MODEL(store), &iter)) {
+      gtk_tree_model_get(GTK_TREE_MODEL(store), &iter, 
+			   TAG_COL_DATA, &tag, -1);
+      g_assert(tag);      
+      gtk_list_store_set(store, &iter,
+	 TAG_COL_COLLISION, info_tag_key_collision(tags, tag), -1);
+    }
+  }
+}
+
 static void on_tag_remove(GtkWidget *but, tag_context_t *context) {
   GtkTreeSelection *selection;
   GtkTreeModel     *model;
@@ -79,6 +113,8 @@ static void on_tag_remove(GtkWidget *but, tag_context_t *context) {
     /* and remove from store */
     gtk_list_store_remove(GTK_LIST_STORE(model), &iter);
 
+    update_collisions(context->store, *context->tag);
+
     /* but is NULL if remove has been called due to a cancelled add */
     if(but) {
       /* something has been removed -> now dirty */
@@ -128,6 +164,7 @@ static gboolean tag_edit(tag_context_t *context) {
   gtk_misc_set_alignment(GTK_MISC(label), 1.f, 0.5f);
   gtk_table_attach_defaults(GTK_TABLE(table), 
 		    key = gtk_entry_new(), 1, 2, 0, 1);
+  gtk_entry_set_activates_default(GTK_ENTRY(key), TRUE);
   HILDON_ENTRY_NO_AUTOCAP(key);
 
   gtk_table_attach_defaults(GTK_TABLE(table),
@@ -135,6 +172,7 @@ static gboolean tag_edit(tag_context_t *context) {
   gtk_misc_set_alignment(GTK_MISC(label), 1.f, 0.5f);
   gtk_table_attach_defaults(GTK_TABLE(table), 
 		    value = gtk_entry_new(), 1, 2, 1, 2);
+  gtk_entry_set_activates_default(GTK_ENTRY(value), TRUE);
   HILDON_ENTRY_NO_AUTOCAP(value);
 
   gtk_entry_set_text(GTK_ENTRY(key), tag->key);
@@ -156,6 +194,9 @@ static gboolean tag_edit(tag_context_t *context) {
 		       -1);
 
     gtk_widget_destroy(dialog);
+
+    /* update collisions for all entries */
+    update_collisions(context->store, *context->tag);
     return TRUE;
   }
 
@@ -168,6 +209,29 @@ static void on_tag_edit(GtkWidget *button, tag_context_t *context) {
     map_item_set_flags(&context->appdata->map->selected, OSM_FLAG_DIRTY, 0);
 }
 
+static void on_tag_last(GtkWidget *button, tag_context_t *context) {
+  static const char *type_name[] = { "illegal", "node", "way", "relation" };
+
+  if(yes_no_f(context->dialog, 
+	      context->appdata, MISC_AGAIN_ID_OVERWRITE_TAGS, 0,
+	      _("Overwrite tags?"),
+	      _("This will overwrite all tags of this %s with the "
+		"ones from the %s selected last.\n\n"
+		"Do you really want this?"),
+	      type_name[context->type], type_name[context->type])) {
+
+    osm_tags_free(*context->tag);
+
+    if(context->type == NODE)
+      *context->tag = osm_tags_copy(context->appdata->map->last_node_tags, TRUE);
+    else
+      *context->tag = osm_tags_copy(context->appdata->map->last_way_tags, TRUE);
+
+    info_tags_replace(context);
+    map_item_set_flags(&context->appdata->map->selected, OSM_FLAG_DIRTY, 0);
+  }
+}
+
 static void on_tag_add(GtkWidget *button, tag_context_t *context) {
   /* search end of tag chain */
   tag_t **tag = context->tag;
@@ -191,6 +255,7 @@ static void on_tag_add(GtkWidget *button, tag_context_t *context) {
   gtk_list_store_set(context->store, &iter,
 		     TAG_COL_KEY, (*tag)->key,
 		     TAG_COL_VALUE, (*tag)->value,
+		     TAG_COL_COLLISION, FALSE,
 		     TAG_COL_DATA, *tag,
 		     -1);
 
@@ -215,10 +280,11 @@ void info_tags_replace(tag_context_t *context) {
   while(tag) {
     gtk_list_store_append(context->store, &iter);
     gtk_list_store_set(context->store, &iter,
-		       TAG_COL_KEY, tag->key,
-		       TAG_COL_VALUE, tag->value,
-		       TAG_COL_DATA, tag,
-		       -1);
+	       TAG_COL_KEY, tag->key,
+	       TAG_COL_VALUE, tag->value,
+	       TAG_COL_COLLISION, info_tag_key_collision(*context->tag, tag),
+	       TAG_COL_DATA, tag,
+	       -1);
     tag = tag->next;
   }
 }
@@ -234,23 +300,33 @@ static GtkWidget *tag_widget(tag_context_t *context) {
 
   /* --- "Key" column --- */
   GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
-  g_object_set(renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL );
+  g_object_set(renderer, 
+	       "ellipsize", PANGO_ELLIPSIZE_END, 
+	       "background", "red", 
+	       NULL );
   GtkTreeViewColumn *column = gtk_tree_view_column_new_with_attributes(
-		 _("Key"), renderer, "text", TAG_COL_KEY, NULL);
+		 _("Key"), renderer, 
+		 "text", TAG_COL_KEY, 
+		 "background-set", TAG_COL_COLLISION,
+		 NULL);
   gtk_tree_view_column_set_expand(column, TRUE);
   gtk_tree_view_insert_column(GTK_TREE_VIEW(context->view), column, -1);
 
   /* --- "Value" column --- */
   renderer = gtk_cell_renderer_text_new();
-  g_object_set(renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL );
+  g_object_set(renderer, 
+	       "ellipsize", PANGO_ELLIPSIZE_END, 
+	       NULL );
   column = gtk_tree_view_column_new_with_attributes(
-		 _("Value"), renderer, "text", TAG_COL_VALUE, NULL);
+		 _("Value"), renderer, 
+		 "text", TAG_COL_VALUE, 
+		 NULL);
   gtk_tree_view_column_set_expand(column, TRUE);
   gtk_tree_view_insert_column(GTK_TREE_VIEW(context->view), column, -1);
 
   /* build and fill the store */
   context->store = gtk_list_store_new(TAG_NUM_COLS, 
-		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_POINTER);
+		G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN, G_TYPE_POINTER);
 
   gtk_tree_view_set_model(GTK_TREE_VIEW(context->view), 
 			  GTK_TREE_MODEL(context->store));
@@ -261,10 +337,11 @@ static GtkWidget *tag_widget(tag_context_t *context) {
     /* Append a row and fill in some data */
     gtk_list_store_append(context->store, &iter);
     gtk_list_store_set(context->store, &iter,
-		       TAG_COL_KEY, tag->key,
-		       TAG_COL_VALUE, tag->value,
-		       TAG_COL_DATA, tag,
-		       -1);
+	       TAG_COL_KEY, tag->key,
+	       TAG_COL_VALUE, tag->value,
+	       TAG_COL_COLLISION, info_tag_key_collision(*context->tag, tag),
+	       TAG_COL_DATA, tag,
+	       -1);
     tag = tag->next;
   }
   
@@ -284,6 +361,21 @@ static GtkWidget *tag_widget(tag_context_t *context) {
 
   GtkWidget *hbox = gtk_hbox_new(TRUE,3);
 
+  GtkWidget *but_last = gtk_button_new_with_label(_("Last..."));
+  gtk_box_pack_start_defaults(GTK_BOX(hbox), but_last);
+
+  /* disable if no appropriate "last" tags have been stored or if the */
+  /* selected item isn't a node or way */
+  if(((context->type == NODE) && 
+      (!context->appdata->map->last_node_tags)) ||
+     ((context->type == WAY) && 
+      (!context->appdata->map->last_way_tags)) ||
+     ((context->type != NODE) && (context->type != WAY)))
+    gtk_widget_set_sensitive(but_last, FALSE);
+
+  gtk_signal_connect(GTK_OBJECT(but_last), "clicked",
+  		     GTK_SIGNAL_FUNC(on_tag_last), context);
+
   GtkWidget *presets = josm_presets_select(context->appdata, context);
   gtk_box_pack_start_defaults(GTK_BOX(hbox), presets);
 
@@ -315,21 +407,25 @@ void info_dialog(appdata_t *appdata) {
   user_t *user = NULL;
   char *str = NULL;
   time_t stime = 0;
+  tag_t *work_copy = NULL;
 
   context->appdata = appdata;
+  context->tag = &work_copy;
 
   switch(appdata->map->selected.type) {
   case MAP_TYPE_NODE:
     str = g_strdup_printf(_("Node #%ld"), appdata->map->selected.node->id);
     user = appdata->map->selected.node->user;
-    context->tag = &appdata->map->selected.node->tag;
+    work_copy = osm_tags_copy(appdata->map->selected.node->tag, FALSE);
     stime = appdata->map->selected.node->time;
+    context->type = NODE;
     break;
   case MAP_TYPE_WAY:
     str = g_strdup_printf(_("Way #%ld"), appdata->map->selected.way->id);
     user = appdata->map->selected.way->user;
-    context->tag = &appdata->map->selected.way->tag;
+    work_copy = osm_tags_copy(appdata->map->selected.way->tag, FALSE);
     stime = appdata->map->selected.way->time;
+    context->type = WAY;
     break;
   default:
     g_assert((appdata->map->selected.type == MAP_TYPE_NODE) ||
@@ -339,11 +435,13 @@ void info_dialog(appdata_t *appdata) {
 
   context->dialog = gtk_dialog_new_with_buttons(str,
 	  GTK_WINDOW(appdata->window), GTK_DIALOG_MODAL,
-	  GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);
+	  GTK_STOCK_OK, GTK_RESPONSE_ACCEPT, 
+	  GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL, 
+	  NULL);
   g_free(str);
 
   gtk_dialog_set_default_response(GTK_DIALOG(context->dialog), 
-				  GTK_RESPONSE_CLOSE);
+				  GTK_RESPONSE_ACCEPT);
 
   /* making the dialog a little wider makes it less "crowded" */
 #ifdef USE_HILDON
@@ -389,7 +487,7 @@ void info_dialog(appdata_t *appdata) {
     gtk_table_attach_defaults(GTK_TABLE(table),  label, 0, 1, 1, 2);
     g_free(nodes_str);
 
-    if(appdata->map->selected.way->draw_flags & OSM_DRAW_FLAGS_POLYGON) 
+    if(appdata->map->selected.way->draw_flags & OSM_DRAW_FLAG_POLYGON) 
       label = gtk_label_new("Closed way (polygon)");
     else
       label = gtk_label_new("Open way (line)");
@@ -409,10 +507,34 @@ void info_dialog(appdata_t *appdata) {
   /* ----------------------------------- */
 
   gtk_widget_show_all(context->dialog);
-  gtk_dialog_run(GTK_DIALOG(context->dialog));
-  gtk_widget_destroy(context->dialog);
+  if(gtk_dialog_run(GTK_DIALOG(context->dialog)) == GTK_RESPONSE_ACCEPT) {
+    gtk_widget_destroy(context->dialog);
 
-  /* since nodes being parts of ways but with no tags are invisible, */
-  /* the result of editing them may have changed their visibility */
-  map_item_redraw(appdata, &appdata->map->selected);
+    /* replace original tags with work copy */
+    switch(appdata->map->selected.type) {
+
+    case MAP_TYPE_NODE:
+      osm_tags_free(appdata->map->selected.node->tag);
+      appdata->map->selected.node->tag = osm_tags_copy(work_copy, TRUE);
+      break;
+
+    case MAP_TYPE_WAY:
+      osm_tags_free(appdata->map->selected.way->tag);
+      appdata->map->selected.way->tag = osm_tags_copy(work_copy, TRUE);
+      break;
+
+    default:
+      break;
+    }
+
+    /* since nodes being parts of ways but with no tags are invisible, */
+    /* the result of editing them may have changed their visibility */
+    map_item_redraw(appdata, &appdata->map->selected);
+
+  } else {
+    gtk_widget_destroy(context->dialog);
+    osm_tags_free(work_copy);
+  }
+
+  g_free(context);
 }
