@@ -25,11 +25,13 @@ typedef struct {
   GtkWidget *dialog, *view;
   GtkListStore *store;
   GtkWidget *but_add, *but_remove, *but_edit;
+  type_t type;
   tag_t **tag;
 } tag_context_t;
 
 
 void info_dialog(appdata_t *appdata);
 void info_tags_replace(tag_context_t *context);
+gboolean info_tag_key_collision(tag_t *tags, tag_t *tag);
 
 #endif // INFO_H
