@@ -23,7 +23,7 @@
 
 #undef DESTROY_WAIT_FOR_GTK
 
-static void map_way_color(potlatch_conf_t *potlatch, way_t *way) {
+void map_way_color(potlatch_conf_t *potlatch, way_t *way) {
   way->color = way->fill_color = MAP_COLOR_NONE;
   way->draw_flags = 0;
 
@@ -124,12 +124,12 @@ static void map_statusbar(map_t *map, map_item_t *map_item) {
   g_free(str);
 }
 
-static void map_outside_error(appdata_t *appdata) {
+void map_outside_error(appdata_t *appdata) {
   errorf(GTK_WIDGET(appdata->window), 
 	 _("Items must not be placed outside the working area!"));
 }
 
-static void map_item_chain_destroy(map_item_chain_t **chainP) {
+void map_item_chain_destroy(map_item_chain_t **chainP) {
   if(!*chainP) {
     printf("nothing to destroy!\n");
     return;
@@ -206,7 +206,7 @@ static void map_node_select(appdata_t *appdata, node_t *node) {
   }
 }
 
-static void map_way_select(appdata_t *appdata, way_t *way) {
+void map_way_select(appdata_t *appdata, way_t *way) {
   map_t *map = appdata->map;
   map_item_t *map_item = &map->selected;
   
@@ -287,7 +287,7 @@ static void map_item_select(appdata_t *appdata, map_item_t *map_item) {
   }
 }
 
-static void map_item_deselect(appdata_t *appdata) {
+void map_item_deselect(appdata_t *appdata) {
 
   /* save tags for "last" function in info dialog */
   if(appdata->map->selected.type == MAP_TYPE_NODE) {
@@ -434,7 +434,7 @@ void map_show_node(map_t *map, node_t *node) {
   map_node_new(map, node, S_RADIUS, 0, NODE_WAY_COLOR, 0);
 }
 
-static void map_way_draw(map_t *map, way_t *way) {
+void map_way_draw(map_t *map, way_t *way) {
 
   /* don't draw a way that's not there anymore */
   if(way->flags & (OSM_FLAG_DELETED | OSM_FLAG_HIDDEN))
@@ -472,7 +472,7 @@ static void map_way_draw(map_t *map, way_t *way) {
   }
 }
 
-static void map_node_draw(map_t *map, node_t *node) {
+void map_node_draw(map_t *map, node_t *node) {
   /* don't draw a node that's not there anymore */
   if(node->flags & OSM_FLAG_DELETED)
     return;
@@ -619,7 +619,7 @@ static gint map_destroy_event(GtkWidget *widget, gpointer data) {
 }
 
 /* get the item at position x, y */
-static map_item_t *map_item_at(map_t *map, gint x, gint y) {	  
+map_item_t *map_item_at(map_t *map, gint x, gint y) {	  
   printf("map check at %d/%d\n", x, y);
 
   canvas_window2world(map->canvas, x, y, &x, &y);
@@ -771,9 +771,14 @@ static void map_do_scroll_step(map_t *map, gint x, gint y) {
 }
 
 
-static gboolean item_is_selected_node(map_t *map, map_item_t *map_item) {
+gboolean map_item_is_selected_node(map_t *map, map_item_t *map_item) {
   printf("check if item is a selected node\n");
 
+  if(!map_item) {
+    printf("  no item requested\n");
+    return FALSE;
+  }
+
   if(map->selected.type == MAP_TYPE_ILLEGAL) {
     printf("  nothing is selected\n");
     return FALSE;
@@ -820,7 +825,7 @@ static gboolean item_is_selected_node(map_t *map, map_item_t *map_item) {
 
 /* return true if the item given is the currenly selected way */
 /* also return false if nothing is selected or the selection is no way */
-static gboolean item_is_selected_way(map_t *map, map_item_t *map_item) {
+gboolean map_item_is_selected_way(map_t *map, map_item_t *map_item) {
   printf("check if item is the selected way\n");
 
   if(!map_item) {
@@ -867,7 +872,7 @@ void map_highlight_refresh(appdata_t *appdata) {
   map_item_select(appdata, &old);
 }
 
-static void map_way_delete(appdata_t *appdata, way_t *way) {
+void map_way_delete(appdata_t *appdata, way_t *way) {
   printf("deleting way #%ld from map and osm\n", way->id);
 
   /* remove it visually from the screen */
@@ -1185,22 +1190,9 @@ static void map_node_move(appdata_t *appdata, map_item_t *map_item,
   
   /* now update the visual representation of the node */
   
-  /* xyz */
   map_item_chain_destroy(&node->map_item_chain);
   map_node_draw(map, node);
 
-#if 0
-  canvas_item_t *item = map_item->item;
-  if(!node->ways) 
-    canvas_item_set_pos(item, &node->lpos, RADIUS);
-  else {
-    /* only move item if it's not the highlight. highlight will be */
-    /* redrawed entirely lateron */
-    if(!map_item->highlight)
-      canvas_item_set_pos(item, &node->lpos, S_RADIUS);
-  }
-#endif
-
   /* visually update ways, node is part of */
   way_t *way = osm->way;
   while(way) {
@@ -1333,27 +1325,25 @@ static void map_button_press(map_t *map, gint x, gint y) {
   map->pen_down.on_selected_node = FALSE;
   if(map->pen_down.on_item)
     map->pen_down.on_selected_node = 
-      item_is_selected_node(map, map->pen_down.on_item);
+      map_item_is_selected_node(map, map->pen_down.on_item);
 
   /* button press */
   switch(map->action.type) {
 
   case MAP_ACTION_WAY_NODE_ADD:
+    map_edit_way_node_add_highlight(map, map->pen_down.on_item, x, y);
+    break;
+
   case MAP_ACTION_WAY_CUT:
-    if(item_is_selected_way(map, map->pen_down.on_item)) {
-      gint nx, ny;
-      canvas_window2world(map->canvas, x, y, &nx, &ny);
-      if(canvas_item_get_segment(map->pen_down.on_item->item, nx, ny) >= 0)
-	map_hl_cursor_draw(map, x, y);
-    }
+    map_edit_way_cut_highlight(map, map->pen_down.on_item, x, y);
     break;
 
   case MAP_ACTION_NODE_ADD:
-    map_hl_cursor_draw(map, x, y);
+    map_hl_cursor_draw(map, x, y, FALSE, RADIUS);
     break;
 
   case MAP_ACTION_WAY_ADD:
-    map_hl_cursor_draw(map, x, y);
+    map_hl_cursor_draw(map, x, y, FALSE, RADIUS);
     map_touchnode_update(map->appdata, x, y);
     break;
 
@@ -1463,248 +1453,21 @@ static void map_button_release(map_t *map, gint x, gint y) {
     printf("released after WAY ADD\n");
     map_hl_cursor_clear(map);
 
-    /* convert mouse position to canvas (world) position */
-    canvas_window2world(map->canvas, x, y, &x, &y);
-
-    /* check if this was a double click. This is the case if */
-    /* the last node placed is less than 5 pixels from the current */
-    /* position */
-    node_t *lnode = osm_way_get_last_node(map->action.way);
-    if(lnode && (map->state->zoom * sqrt((lnode->lpos.x-x)*(lnode->lpos.x-x)+ 
-				  (lnode->lpos.y-y)*(lnode->lpos.y-y))) < 5) {
-      printf("detected double click -> simulate ok click\n");
-      map_hl_touchnode_clear(map);
-      map_action_ok(map->appdata);
-    } else {
-      
-      /* use the existing node if one was touched */
-      node = map_hl_touchnode_get_node(map);
-      if(node) {
-	printf("  re-using node #%ld\n", node->id);
-	map_hl_touchnode_clear(map);
-      } else {
-	if(!osm_position_within_bounds(map->appdata->osm, x, y)) 
-	  map_outside_error(map->appdata);
-	else 
-	  node = osm_node_new(map->appdata->osm, x, y);
-      }
-      
-      if(node) {
-	g_assert(map->action.way);
-	osm_way_append_node(map->action.way, node);
-	
-	switch(osm_node_chain_length(map->action.way->node_chain)) {
-	case 1:
-	  /* replace "place first node..." message */
-	  statusbar_set(map->appdata, _("Place next node of way"), FALSE);
-	  break;
-	  
-	case 2:
-	  /* two nodes are enough for a valid way */
-	  icon_bar_map_cancel_ok(map->appdata, TRUE, TRUE);
-	  break;
-	  
-	default:
-	  break;
-	}
-	
-	/* remove prior version of this way */
-	map_item_chain_destroy(&map->action.way->map_item_chain);
-	
-	/* draw current way */
-	map_way_color(map->appdata->potlatch, map->action.way);
-	map_way_draw(map, map->action.way);
-      }
-    }
+    map_edit_way_add_segment(map, x, y);
     break;
 
   case MAP_ACTION_WAY_NODE_ADD:
     printf("released after WAY NODE ADD\n");
     map_hl_cursor_clear(map);
 
-    /* check if we are still hovering above the selected way */
-    map_item_t *item = map_item_at(map, x, y);
-    if(item && item_is_selected_way(map, item)) {
-      /* convert mouse position to canvas (world) position */
-      canvas_window2world(map->canvas, x, y, &x, &y);
-      gint insert_after = canvas_item_get_segment(item->item, x, y)+1;
-      if(insert_after > 0) {
-	/* create new node */
-	node_t* node = osm_node_new(map->appdata->osm, x, y);
-	osm_node_attach(map->appdata->osm, node);
-
-	/* insert it into ways chain of nodes */
-	way_t *way = item->way;
-
-	/* search correct position */
-	node_chain_t **chain = &way->node_chain;
-	while(insert_after--) {
-	  g_assert(*chain);
-	  chain = &(*chain)->next;
-	}
-
-	/* and actually do the insertion */
-	node_chain_t *new_chain_item = g_new0(node_chain_t, 1);
-	new_chain_item->node = node;
-	new_chain_item->next = *chain;
-	*chain = new_chain_item;
-
-	/* clear selection */
-	map_item_deselect(map->appdata);
-
-	/* remove prior version of this way */
-	map_item_chain_destroy(&way->map_item_chain);
-
-	/* draw the updated way */
-	map_way_draw(map, way);
-
-	/* remember that this node is contained in one way */
-	node->ways=1;
-
-	/* and that the way needs to be uploaded */
-	way->flags |= OSM_FLAG_DIRTY;
-
-	/* put gui into idle state */
-	map_action_set(map->appdata, MAP_ACTION_IDLE);
-
-	/* and redo it */
-       	map_way_select(map->appdata, way);
-      }
-    }
+    map_edit_way_node_add(map, x, y);
     break;
 
   case MAP_ACTION_WAY_CUT:
     printf("released after WAY CUT\n");
     map_hl_cursor_clear(map);
 
-    /* check if we are still hovering above the selected way */
-    item = map_item_at(map, x, y);
-    if(item && item_is_selected_way(map, item)) {
-      /* convert mouse position to canvas (world) position */
-      canvas_window2world(map->canvas, x, y, &x, &y);
-      gint insert_after = canvas_item_get_segment(item->item, x, y);
-      if(insert_after >= 0) {
-	/* create a duplicate of the currently selected way */
-	way_t *way = item->way;
-	way_t *new = osm_way_new();
-
-	/* if this is a closed way, reorder (rotate) it, so the */
-	/* place to cut is adjecent to the begin/end of the way. */
-	/* this prevents a cut polygon to be split into two ways */
-	g_assert(way->node_chain);
-	if(way->node_chain->node == osm_way_get_last_node(way)) {
-	  printf("CLOSED WAY -> rotate by %d\n", insert_after);
-	  osm_way_rotate(way, insert_after);
-	  insert_after = 0;
-	}
-
-	/* ------------  copy all tags ------------- */
-	new->tag = osm_tags_copy(way->tag, TRUE);
-
-	/* ---- transfer relation membership from way to new ----- */
-	relation_chain_t *rchain = osm_way_to_relation(map->appdata->osm, way);
-
-	while(rchain) {
-	  relation_chain_t *next = rchain->next;
-	  printf("way is part of relation #%ld\n", rchain->relation->id);
-
-	  /* make new member of the same relation */
-
-	  /* walk member chain. save role of way if its being found. */
-	  member_t **member = &rchain->relation->member;
-	  char *role = NULL;
-	  while(*member) { 
-	    /* save role of way */
-	    if(((*member)->type == WAY) && ((*member)->way == way))
-	      role = (*member)->role;
-	    member = &(*member)->next;
-	  }
-
-	  printf("  adding new to relation\n");
-	  *member = g_new0(member_t, 1);
-	  (*member)->type = WAY;
-	  (*member)->way = new;
-	  if(role) (*member)->role = g_strdup(role);
-	  member = &(*member)->next;
-	  
-	  rchain->relation->flags |= OSM_FLAG_DIRTY;
-
-	  g_free(rchain);
-	  rchain = next;
-	}
-
-	/* move parts of node_chain to the new way */
-	node_chain_t *chain = way->node_chain;
-	while(insert_after--) {
-	  g_assert(chain);
-	  chain = chain->next;
-	}
-	 
-	/* attach remaining nodes to new way */
-	new->node_chain = chain->next;
-
-	/* terminate remainig chain on old way */
-	chain->next = NULL;
-
-	/* now move the way itself into the main data structure */
-	osm_way_attach(map->appdata->osm, new);
-	
-	/* clear selection */
-	map_item_deselect(map->appdata);
-
-	/* remove prior version of this way */
-	printf("remove visible version of way #%ld\n", way->id);
-	map_item_chain_destroy(&way->map_item_chain);
-
-	/* swap chains of the old way is to be destroyed due to a lack */
-	/* of nodes */
-	if(osm_way_number_of_nodes(way) < 2) {
-	  printf("swapping ways to avoid destruction of original way\n");
-	  node_chain_t *tmp = way->node_chain;
-	  way->node_chain = new->node_chain;
-	  new->node_chain = tmp;
-	}
-
-	/* the way may still only consist of a single node. */
-	/* remove it then */
-	if(osm_way_number_of_nodes(way) < 2) {
-	  printf("original way has less than 2 nodes left, deleting it\n");
-	  map_way_delete(map->appdata, way);
-	  item = NULL;
-	} else {
-	  printf("original way still has %d nodes\n",
-		 osm_way_number_of_nodes(way));
-
-	  /* draw the updated old way */
-	  map_way_color(map->appdata->potlatch, way);
-	  map_way_draw(map, way);
-
-	  /* remember that the way needs to be uploaded */
-	  way->flags |= OSM_FLAG_DIRTY;
-	}
-
-	if(osm_way_number_of_nodes(new) < 2) {
-	  printf("new way has less than 2 nodes, deleting it\n");
-	  map_way_delete(map->appdata, new);
-	  new = NULL;
-	} else {
-
-	  /* colorize the new way before drawing */
-	  map_way_color(map->appdata->potlatch, new);
-	  map_way_draw(map, new);
-	}
-
-	/* put gui into idle state */
-	map_action_set(map->appdata, MAP_ACTION_IDLE);      
-
-	/* and redo selection if way still exists */
-	if(item) 
-	  map_way_select(map->appdata, way);
-	else if(new)
-	  map_way_select(map->appdata, new);
-      }
-    }
-
+    map_edit_way_cut(map, x, y);
     break;
 
 
@@ -1777,31 +1540,31 @@ static gboolean map_motion_notify_event(GtkWidget *widget,
       if(!map->pen_down.on_selected_node)
 	map_do_scroll(map, x, y);
       else {
-	map_hl_cursor_draw(map, x, y);
+	map_hl_cursor_draw(map, x, y, FALSE, RADIUS);
 	map_touchnode_update(appdata, x, y);
       }
     }
     break;
     
   case MAP_ACTION_NODE_ADD:
-    map_hl_cursor_draw(map, x, y);
+    map_hl_cursor_draw(map, x, y, FALSE, RADIUS);
     break;
     
   case MAP_ACTION_WAY_ADD:
-    map_hl_cursor_draw(map, x, y);
+    map_hl_cursor_draw(map, x, y, FALSE, RADIUS);
     map_touchnode_update(appdata, x, y);
     break;
 
   case MAP_ACTION_WAY_NODE_ADD:
-  case MAP_ACTION_WAY_CUT:
     map_hl_cursor_clear(map);
     map_item_t *item = map_item_at(map, x, y);
-    if(item && item_is_selected_way(map, item)) {
-      gint nx, ny;
-      canvas_window2world(map->canvas, x, y, &nx, &ny);
-      if(canvas_item_get_segment(item->item, nx, ny) >= 0)
-	map_hl_cursor_draw(map, x, y);
-    }
+    if(item) map_edit_way_node_add_highlight(map, item, x, y);
+    break;
+
+  case MAP_ACTION_WAY_CUT:
+    map_hl_cursor_clear(map);
+    item = map_item_at(map, x, y);
+    if(item) map_edit_way_cut_highlight(map, item, x, y);
     break;
 
   default:
@@ -2087,33 +1850,7 @@ void map_action_cancel(appdata_t *appdata) {
 
   switch(map->action.type) {
   case MAP_ACTION_WAY_ADD:
-    printf("  removing temporary way\n");
-    g_assert(map->action.way);
-
-    /* remove all nodes that have been created for this way */
-    /* (their way count will be 0 after removing the way) */
-    node_chain_t *chain = map->action.way->node_chain;
-    while(chain) {
-      node_chain_t *next = chain->next;
-
-      printf("    node #%ld (used by %d)\n", 
-	     chain->node->id, chain->node->ways);
-
-      chain->node->ways--;
-      if(!chain->node->ways && (chain->node->id == ID_ILLEGAL)) {
-	printf("      -> freeing temp node\n");
-	osm_node_free(chain->node);
-      }
-      g_free(chain);
-      chain = next;
-    }
-    map->action.way->node_chain = NULL;
-
-    /* remove ways visual representation */
-    map_item_chain_destroy(&map->action.way->map_item_chain);
-
-    osm_way_free(map->action.way);
-    map->action.way = NULL;
+    map_edit_way_add_cancel(map);
     break;
 
   case MAP_ACTION_BG_ADJUST:
@@ -2138,44 +1875,7 @@ void map_action_ok(appdata_t *appdata) {
 
   switch(map->action.type) {
   case MAP_ACTION_WAY_ADD:
-    g_assert(map->action.way);
-    
-    /* transfer all nodes that have been created for this way */
-    /* into the node chain */
-
-    /* (their way count will be 0 after removing the way) */
-    node_chain_t *chain = map->action.way->node_chain;
-    while(chain) {
-      printf("    node #%ld (used by %d)\n", 
-	     chain->node->id, chain->node->ways);
-
-      /* a node may have been a stand-alone node before, so remove its */
-      /* visual representation as its now drawn as part of the way */
-      /* (if at all) */
-      if(chain->node->id != ID_ILLEGAL) 
-	map_item_chain_destroy(&chain->node->map_item_chain);
-
-      map_node_draw(map, chain->node);
-
-      /* we can be sure that no node gets inserted twice (even if twice in */
-      /* the ways chain) because it gets assigned a non-ID_ILLEGAL id when */
-      /* being moved to the osm node chain */
-      if(chain->node->id == ID_ILLEGAL) 
-	osm_node_attach(appdata->osm, chain->node);
-
-      chain = chain->next;
-    }
-
-    /* now move the way itself into the main data structure */
-    osm_way_attach(appdata->osm, map->action.way);
-
-    map_way_select(appdata, map->action.way);
-
-    map->action.way = NULL;
-
-    /* let the user specify some tags for the new way */
-    info_dialog(appdata);
-
+    map_edit_way_add_ok(map);
     break;
 
   case MAP_ACTION_BG_ADJUST:
