@@ -188,4 +188,17 @@ void map_show_all(appdata_t *appdata);
 
 void map_set_zoom(map_t *map, double zoom, gboolean update_scroll_offsets);
 
+/* various functions required by map_edit */
+gboolean map_item_is_selected_node(map_t *map, map_item_t *map_item);
+gboolean map_item_is_selected_way(map_t *map, map_item_t *map_item);
+void map_item_chain_destroy(map_item_chain_t **chainP);
+map_item_t *map_item_at(map_t *map, gint x, gint y);
+void map_item_deselect(appdata_t *appdata);
+void map_way_delete(appdata_t *appdata, way_t *way);
+void map_way_color(potlatch_conf_t *potlatch, way_t *way);
+void map_way_draw(map_t *map, way_t *way);
+void map_way_select(appdata_t *appdata, way_t *way);
+void map_outside_error(appdata_t *appdata);
+void map_node_draw(map_t *map, node_t *node);
+
 #endif // MAP_H
