@@ -864,7 +864,7 @@ void osm_upload(appdata_t *appdata, osm_t *osm, project_t *project) {
 
       appendf(&context->log, _("Reloading map ...\n"));
       
-      if(!diff_is_clean(appdata->osm)) {
+      if(!diff_is_clean(appdata->osm, FALSE)) {
 	appendf(&context->log, _(">>>>>>>> DIFF IS NOT CLEAN <<<<<<<<\n"));
 	appendf(&context->log, _("Something went wrong during upload,\n"));
 	appendf(&context->log, _("proceed with care!\n"));
