@@ -21,11 +21,12 @@
 #define STATUSBAR_H
 
 typedef struct statusbar_s {
-  GtkWidget *widget;
+  GtkWidget *widget, *eventbox;
   int cid;
 } statusbar_t;
 
-void statusbar_set(appdata_t *appdata, const char *msg);
+void statusbar_set(appdata_t *appdata, const char *msg, gboolean highlight);
 GtkWidget *statusbar_new(appdata_t *appdata);
+void statusbar_highlight(appdata_t *appdata, gboolean highlight);
 
 #endif // STATUSBAR_H
